define([
    "dojo/_base/declare",
    "dojo/_base/array",
    "dojo/_base/lang",
    "dojo/on",
    "dojo/promise/all",
    "dojo/when",
    "dojo/json",
    "dojo/dom-class",
    "dijit/form/RadioButton",
    "dijit/form/CheckBox",
    "dojo/query",
    "dojo/Deferred",
    "thornberry/app/utils"
],function(
    declare,
    arrayUtils,
    lang,
    on,
    all,
    when,
    json,
    domClass,
    RadioButton,
    CheckBox,
    query,
    Deferred,
    utils
) {
    // IMPORTANT: requires that the widget this is mixed into has a form property of type dijit/form/Form 

    var DEFAULT_TEXT_MAX_LENGTH = 50;

    var _EditorMixin = declare(null, {
        disabled: false, 
        modified: false,
        fields: null,
        fieldMap: null,

        value: null,   // external form values (always use get/set)
        history: null, // array of (internal) values at each change, the highest being the most recent
        validationPromise: null, // a promise created when an async validation process begins that resolves when complete

        _modifiedInvalidated: false,
        _fieldStateInvalidated: false,
        _fieldStateUpdateDeferred: null,
        _isSubEditor: false,
        _firstValueString: null,
        _saved: false,

        onChange: function() {},

        conversions: null,
        addConversion: function(names,conversion) {
            // conversion = function(name,value,dir) {return converted} where dir = "set" or "get"
            // Only a single conversion function may be set per field. If multiple are added, only the first is called
            this._setNameFunctionIndex(names,conversion,this.conversions);
        },

        validations: null,
        addValidation: function(names,validation) {
            // names = null for any change, name for single field, or [] for multiple fields
            // validation = function(name,fieldValue,lastFieldValue)
            //      returns either true or false or a promise that resolves to true or false
            // Use addValidation() to set functions that run on user input events that change form values. First, field specific validation runs, and then validation for no field
            // Only a single validation function may be set per field. If multiple are added, only the first is called
            this._addToNameFunctionIndex(names,validation,this.validations);
        },

        logic: null,
        setValueLogic: null,
        addLogic: function(names,logic,onSetValue) {
            // names = null for any change, name for single field, or [] for multiple fields
            // logic = function(name,fieldValue,lastFieldValue)
            // Use addLogic() to set functions that run on user input events that change form values. First, field specific logic runs, and then logic for any change
            // Logic functions are only run after validation and if validation is successful
            // Multiple can be set up per field, and all will be called when that field changes
            this._addToNameFunctionIndex(names,logic,this.logic);
            if (onSetValue) {
                this._addToNameFunctionIndex(names,logic,this.setValueLogic);
            }
        },

        requirements: null,
        addRequirement: function(names,requirement,onSetValue) {
            // names = null for all fields, name for single field, or [] for multiple fields
            // requirement = boolean or function(name,value,lastValue), value is an object with the current internal field values, and lastValue has the set of field values from before the most recent input
            //      return true if the field should be required
            // Use addRequirement() to set functions that run on user input events that change form values. First, field specific requirement logic runs, and then logic for all fields
            // Multiple can be set up per field, and all will be called on each change. If any return true, then the field is required
            this._addToNameFunctionIndex(names,requirement,this.requirements);
        },

        disableds: null,
        addDisabled: function(names,disabled) {
            // names = null for all fields, name for single field, or [] for multiple fields
            // disabled = boolean or function(name,value,lastValue), value is an object with the current (internal) field values, and lastValue has the set of field values from before the most recent input
            //      return true if the field should be disabled
            // Use addDisabled() to set functions that run on user input events that change form values. First, field specific disable logic runs, and then logic for all fields
            // Multiple can be set up per field, and all will be called on each change. If any return true, then the field is disabled
            this._addToNameFunctionIndex(names,disabled,this.disableds);
        },

        _addToNameFunctionIndex: function(names,func,index) {
            var add = function(func,name) {
                if (typeof index[name] === "undefined") {
                    index[name] = [func];
                } else {
                    index[name].push(func);
                }
            };
            if (lang.isArray(names)) {
                arrayUtils.forEach(names,function(name) {
                    add(func,name);
                });
            } else if (names) {
                add(func,names);
            } else {
                add(func,"");
            }
        },

        _setNameFunctionIndex: function(names,func,index) {
            if (lang.isArray(names)) {
                arrayUtils.forEach(names,function(name) {
                    index[name] = func;
                });
            } else if (names) {
                index[names] = func;
            } else {
                index[""] = func;
            }
        },

        setFormValue: function(name,value,triggerChanges,multiple) {
            var t = this;
            var formVal = "";
            if (multiple) {
                formVal = value;
            } else {
                var formVal = {};
                formVal[name] = value;
            }
            if (!triggerChanges) {arrayUtils.forEach(t.fields,function(field) { field._onChangeActive = false; });}
            t.form.set("value",formVal);
            if (!triggerChanges) {arrayUtils.forEach(t.fields,function(field) { field._onChangeActive = true; });}
        },

        onChangeModified: function(value) {},   // hook for whenever the modified status of the editor changes. Called with the new value of modified
        onValidationReject: function(name,value,lastValue,rejectValue) {return false;},  // hook for when an async validation promise is rejected. Called with the name of the field that triggered validation, the new value and the old value. 
            // Return true to accept the value, and false to reject and revert the form.

        constructor: function() {
            var t = this;
            t.fields = [];
            t.fieldMap = {};
            t.conversions = {};
            t.validations = {};
            t.logic = {};
            t.setValueLogic = {};
            t.requirements = {};
            t.disableds = {};
            t.history = [];
            t._fieldStateUpdateDeferred = new Deferred();
        },

        postCreate: function() {
            var t = this;

            t._initForm();

            t.fieldMap = t._getWidgetNameMap();
            for (var name in t.fieldMap) {
                arrayUtils.forEach(t.fieldMap[name],function(widget) {
                    t.fields.push(widget);
                    if (widget.isInstanceOf(RadioButton)) {
                        t.own(on(widget,"click",function() {
                            t.defer(function() {
                                t._onInput(widget.name,widget.value);
                            });
                        }));
                    } else if (widget.isInstanceOf(CheckBox)) {
                        t.own(on(widget,"change",function(value) {t._onInput(widget.name,t.form.get("value")[widget.name]);}));
                    } else {
                        if (widget.isInstanceOf(_EditorMixin)) {
                            widget._isSubEditor = true;
                        }
                        t.own(on(widget,"change",function(value) {t._onInput(widget.name,value);}));
                    }

                    if (widget.declaredClass === "dijit.form.DateTextBox") {
                        widget.own(on(widget,"blur",function(){checkForT(this.id);}));
                        t.addConversion(name,utils.conversions.date);
                        widget.constraints.min = utils.validators.minDate;
                    } else if (widget.declaredClass === "dijit.form.ValidationTextBox") {
                        if (!widget.maxLength) {
                            widget.set("maxLength",DEFAULT_TEXT_MAX_LENGTH);
                        }
                    }

                    t.setUpConversion(name,widget);
                });
            }

            t.inherited(arguments);
        },

        setUpConversion: function(name,widget) {
            var t = this;
            if (widget.editorConversion) {
                if (utils.conversions[widget.editorConversion]) {
                    t.addConversion(name,utils.conversions[widget.editorConversion]);
                }
            } else {
                if (widget.declaredClass === "dijit.form.DateTextBox"){
                    t.addConversion(name,utils.conversions.date);
                } else if (widget.declaredClass === "dijit.form.CurrencyTextBox") {
                    t.addConversion(name,utils.conversions.number);
                } else if (widget.declaredClass === "dijit.form.NumberSpinner") {
                    t.addConversion(name,utils.conversions.number);
                } else if (widget.declaredClass === "dijit.form.CheckBox") {
                    t.addConversion(name,utils.conversions.list);
                } else if (widget.declaredClass === "dojox.form.CheckedMultiSelect") {
                    t.addConversion(name,utils.conversions.list);
                }
            }
        },

        reset: function() {
            this.set("value",null);
        },

        updateModified: function(subClassValue) {
            var t = this;
            var value = t.get("value");
            var firstValue = t.value;
            var modified = subClassValue || !firstValue || false;
            if (!modified) {
                for (field in value) {
                    var first = value[field];
                    var second = firstValue[field];
                    if (typeof first === "object" || typeof second === "object") {
                        if (json.stringify(first) != json.stringify(second)) {
                            modified = true;
                        }
                    } else {
                        if (first != second) {
                            modified = true;
                        }
                    }
                    if (modified) {
                        break;
                    }
                }
            }
            if (modified) {
                t.history.push(t.form.get("value"));
                t.onChange(value);
            }
            t._saved = false;
            t._modifiedInvalidated = false;
            t.set("modified",modified);
        },

        _setValueAttr: function(value) {
            var t = this;
            t.inherited(arguments);
            var firstValueString = json.stringify(value);
            value = value || t.getInitialProps().value;
            if (t._needsRefresh(value)) {
                t._firstValueString = firstValueString;
                arrayUtils.forEach(t.fields,function(field) { field._onChangeActive = false; });
                var intVal = null;
                t._reset();
                if (value) {
                    intVal = t.convert(value,"set");
                    t.form.set("value",intVal);
                }
                intVal = lang.mixin(t.form.get("value"),intVal);
                t.history = [intVal];
                t.value = lang.mixin(t.get("value"),t.convert(intVal,"get"));

                arrayUtils.forEach(t.fields,function(field) { field._onChangeActive = true; });

                // execute "onSetValue" field logic
                for (var name in t.fieldMap) {
                    var logic = [];
                    if (t.setValueLogic[name]) {logic = logic.concat(t.setValueLogic[name]);}
                    if (t.setValueLogic[""]) {logic = logic.concat(t.setValueLogic[""]);}
                    arrayUtils.forEach(logic, function(logic) {logic(name,intVal[name],intVal,true);});
                }
                
                t.set("modified",false);
                t.invalidate("fieldState");
            } else {
                t.invalidate("fieldState");
                /*t._fieldStateUpdateDeferred.resolve();
                t.defer(function() {
                    t._fieldStateUpdateDeferred = new Deferred();
                });*/
            }
        },

        _needsRefresh: function(value) {
            var t = this; var t = this;
            var result = false;
            var valueString = json.stringify(value);
            if (t.modified || t._firstValueString !== valueString || t._saved) {
                result = true;
            }
            return result;
        },

        _getValueAttr: function() {
            return this.convert(this.form.get("value"),"get");
        },

        _setModifiedAttr: function(value) {
            var t = this;
            if (t.modified != value) {
                t.modified = value;
                t.onChangeModified(value);
            }
        },

        _reset: function() {
            var t = this;
            arrayUtils.forEach(t.fields,function(field) {  
                var onChangeActive = field._onChangeActive;
                field._onChangeActive = false;
                if (field.reset) {field.reset();} 
                if (field.declaredClass === "dojox.form.CheckedMultiSelect") {
                    field.set("value",[]);
                    field._updateSelection();
                } else if (field.declaredClass === "dijit.form.DateTextBox") {
                    field.set("value",null);
                } else if (field.declaredClass === "dijit.form.CheckBox") {
                    field.set("checked",false);
                } else if (field.declaredClass === "dijit.form.FilteringSelect") {
                    field._lastValueReported = "";
                }
                field._onChangeActive = onChangeActive;
            });
        },

        _onInput: function(name,value) {
            var t = this;

            // invalid character removal
            if (value && value.toString().match(/(~|\^|`)/g)) {
                value = value.toString().replace(/(~|\^|`)/g,"");
                t.setFormValue(name,value);
            }

            var lastValues = t.getHistory();

            // execute validation here - can return promises
            var validations = [];
            if (t.validations[name]) {validations = validations.concat(t.validations[name]);}
            if (t.validations[""]) {validations = validations.concat(t.validations[""]);}
            var validationPromises = [];
            var valid = arrayUtils.every(validations, function(validation) {
                var result = validation(name,value,lastValues);
                if (result && result.then) {validationPromises.push(result);}
                return !!result;
            }); 
            if (valid && validationPromises.length > 0) {
                valid = all(validationPromises);
                t.set("validationPromise",valid);
            }
            var onComplete = function(valid) {
                if (valid) {
                    // execute field logic, which may result in other field changes
                    var logic = [];
                    if (t.logic[name]) {logic = logic.concat(t.logic[name]);}
                    if (t.logic[""]) {logic = logic.concat(t.logic[""]);}
                    arrayUtils.forEach(logic, function(logic) {logic(name,value,lastValues,false);});

                    // execute requirement checks and disabled checks
                    t.invalidate("fieldState");
                    t.invalidate("modified");
                } else {
                    // wait a moment to make it look smoother
                    setTimeout(function() {t.setFormValue(name,lastValues[name]);},250);
                }
            };

            when(valid,function(result) {
                result = typeof result === "boolean" ? result : arrayUtils.every(result,function(valid) {return valid;});
                onComplete(result);
            },function(result) {
                onComplete(t.onValidationReject(name,value,lastValues,result));
            });
        },

        updateFieldState: function() {
            // Disables and requires fields based on current form state
            var t = this;
            var map = t.fieldMap;
            var value = t.form.get("value");
            var lastValue = t.getHistory();

            for (var name in map) {
                var disabled = t.disabled;
                if (!disabled) {
                    var disableds = [];
                    if (t.disableds[name]) {disableds = disableds.concat(t.disableds[name]);}
                    if (t.disableds[""]) {disableds = disableds.concat(t.disableds[""]);}
                    disabled =  (arrayUtils.some(disableds, function(disabled) {return typeof disabled === "boolean" ? disabled : disabled(name,value,lastValue);}));
                }
                if (disabled != map[name][0].get("disabled") || map[name][0].declaredClass === "dojox.form.CheckedMultiSelect") {
                    arrayUtils.forEach(map[name],function(widget) {widget.set("disabled",disabled);});
                }

                var makePink = false;
                if (!disabled) {
                    var requirements = [];
                    if (t.requirements[name]) {requirements = requirements.concat(t.requirements[name]);}
                    if (t.requirements[""]) {requirements = requirements.concat(t.requirements[""]);}
                    var required = (arrayUtils.some(requirements, function(requirement) {return typeof requirement === "boolean" ? requirement : requirement(name,value,lastValue);}));
                    var val = value[name];
                    var isNull = (val === "" || /^\s+$/.test(val) || /*NaN check*/ val != val || val === null || typeof val === "undefined" || (typeof val === "object" && val.length === 0));
                    makePink = (required && isNull);
                }
                arrayUtils.forEach(map[name],function(widget) {
                    var node = (widget.reqBkgdNode ? widget.reqBkgdNode : widget.domNode);
                    domClass.toggle(node,"pinkBackground",makePink);
                });

                arrayUtils.forEach(map[name],function(widget) {
                    if (widget.isInstanceOf(_EditorMixin)) {
                        widget.updateFieldState();
                    }
                });
            }
            t._fieldStateInvalidated = false;
            t._fieldStateUpdateDeferred.resolve();
            t._fieldStateUpdateDeferred = new Deferred();
        },

        _getWidgetNameMap: function() {
            var map = {};
			arrayUtils.forEach(this.form._getDescendantFormWidgets(), function(widget){
				if(!widget.name){ return; }
				var entry = map[widget.name] || (map[widget.name] = [] );
				entry.push(widget);
            });
            return map;
        },

        invalidate: function(prop) {
            var t = this;
            if (prop === "fieldState" && t._isSubEditor) {
                return;
            }
            if (!t["_"+prop+"Invalidated"]) {
                t["_"+prop+"Invalidated"] = true;
                t.defer("update"+prop.charAt(0).toUpperCase()+prop.slice(1));
            }
        },

        getHistory: function(index) {
            var t = this;
            index = typeof index === "undefined" ? t.history.length - 1 : index;
            return (index < t.history.length ? t.history[index] : null);
        },

        getField: function(name,index) {
            index = index || 0;
            return this.fieldMap[name][index];
        },

        convert: function(value,dir) {
            var t = this;
            value = lang.mixin({},value);
            for (field in t.conversions) {
                value[field] = t.conversions[field](field,value[field],dir);
            }
            return value;
        },

        checkRequired: function() {
            return query(".pinkBackground",this.domNode).length === 0;
        },

        checkValid: function() {
            return this.form.validate();
        },

        _setDisabledAttr: function(value) {
            var t = this;
            if (value !== t.disabled) {
                t.invalidate("fieldState");
            }
            t.disabled = value;
        },

        onSaveComplete: function() {
            var t = this;
            t.modified = false;
            t._saved = true;
        },

        _initForm: function() {
            this.form._getValueAttr = function(){
                // copied from _FormMixin
                // changes:
                //      - still get the value from disabled widgets
                var obj = { };
                arrayUtils.forEach(this._getDescendantFormWidgets(), function(widget){
                    var name = widget.name;
                    if(!name){ return; }
    
                    // Single value widget (checkbox, radio, or plain <input> type widget)
                    var value = widget.get('value');
    
                    // Store widget's value(s) as a scalar, except for checkboxes which are automatically arrays
                    if(typeof widget.checked == 'boolean'){
                        if(/Radio/.test(widget.declaredClass)){
                            // radio button
                            if(value !== false){
                                lang.setObject(name, value, obj);
                            }else{
                                // give radio widgets a default of null
                                value = lang.getObject(name, false, obj);
                                if(value === undefined){
                                    lang.setObject(name, null, obj);
                                }
                            }
                        }else{
                            // checkbox/toggle button
                            var ary=lang.getObject(name, false, obj);
                            if(!ary){
                                ary=[];
                                lang.setObject(name, ary, obj);
                            }
                            if(value !== false){
                                ary.push(value);
                            }
                        }
                    }else{
                        var prev=lang.getObject(name, false, obj);
                        if(typeof prev != "undefined"){
                            if(lang.isArray(prev)){
                                prev.push(value);
                            }else{
                                lang.setObject(name, [prev, value], obj);
                            }
                        }else{
                            // unique name
                            lang.setObject(name, value, obj);
                        }
                    }
                });
                return obj;
            };
        }
    });
    return _EditorMixin;
});